﻿using SampleBrowser.Samples.GridView.RowReorder;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Interactivity;
using Telerik.Windows.Controls;
using Telerik.Windows.Controls.DragDrop;
using Telerik.Windows.Controls.GridView;
using Telerik.Windows.Controls.TreeView;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Diagnostics;
using SampleBrowser.Shared;
using System.Globalization;

namespace Samples.GridView.RowReorder
{
    /// <summary>
    /// Provides row (item) reoder capabilities to RadGridView
    /// </summary>
    public partial class RowReorderBehavior : Behavior<RadGridView>
    {
        private const string DropPositionFeedbackElementName = "DragBetweenItemsFeedback";

        private TreeViewDragCue dragCue;
        private object currentDropItem;
        private DropPosition currentDropPosition;
        private ContentPresenter dropPositionFeedbackPresenter;
        private Grid dropPositionFeedbackPresenterHost;

        public event EventHandler<BeginningDragEventArgs> BeginningDrag;
        public event EventHandler<DragStartedEventArgs> DragStarted;
        public event EventHandler<ReorderingEventArgs> Reordering;
        public event EventHandler<ReorderedEventArgs> Reordered;

        public RowReorderBehavior()
        {
            this.dropPositionFeedbackPresenter = new ContentPresenter();
            this.dropPositionFeedbackPresenter.Name = DropPositionFeedbackElementName;
            this.dropPositionFeedbackPresenter.HorizontalAlignment = HorizontalAlignment.Left;
            this.dropPositionFeedbackPresenter.VerticalAlignment = VerticalAlignment.Top;
            this.dropPositionFeedbackPresenter.RenderTransformOrigin = new Point(0.5, 0.5);
        }

        protected override void OnAttached()
        {
            base.OnAttached();

            this.AssociatedObject.RowLoaded += new EventHandler<Telerik.Windows.Controls.GridView.RowLoadedEventArgs>(AssociatedObject_RowLoaded);
            this.AssociatedObject.DataLoaded += new EventHandler<EventArgs>(AssociatedObject_DataLoaded);
            this.AssociatedObject.SetValue(RadDragAndDropManager.AllowDropProperty, true);
            this.SubscribeToDragDropEvents();
        }

        void AssociatedObject_DataLoaded(object sender, EventArgs e)
        {
            this.AssociatedObject.DataLoaded -= new EventHandler<EventArgs>(AssociatedObject_DataLoaded);
            this.AttachDropPositionFeedback();
        }

        protected override void OnDetaching()
        {
            base.OnDetaching();

            this.AssociatedObject.RowLoaded -= new EventHandler<Telerik.Windows.Controls.GridView.RowLoadedEventArgs>(AssociatedObject_RowLoaded);
            this.AssociatedObject.SetValue(RadDragAndDropManager.AllowDropProperty, false);
            this.UnsubscribeFromDragDropEvents();
            this.DetachDropPositionFeedback();
        }

        private void DetachDropPositionFeedback()
        {
            if (this.IsDropPositionFeedbackAvailable())
                this.dropPositionFeedbackPresenterHost.Children.Remove(this.dropPositionFeedbackPresenter);
        }

        private void AttachDropPositionFeedback()
        {
            this.dropPositionFeedbackPresenterHost = this.AssociatedObject.ChildrenOfType<Grid>().FirstOrDefault();

            if (this.dropPositionFeedbackPresenterHost != null)
            {
                this.dropPositionFeedbackPresenter.Content = CreateDefaultDropPositionFeedback();
                this.dropPositionFeedbackPresenterHost.Children.Add(this.dropPositionFeedbackPresenter);
            }

            this.HideDropPositionFeedbackPresenter();
        }

        private static UIElement CreateDefaultDropPositionFeedback()
        {
            Grid grid = new Grid()
            {
                Height = 8,
                HorizontalAlignment = HorizontalAlignment.Stretch,
                IsHitTestVisible = false,
                VerticalAlignment = VerticalAlignment.Stretch
            };

            grid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(8) });
            grid.ColumnDefinitions.Add(new ColumnDefinition());

            Ellipse ellipse = new Ellipse()
            {
                Stroke = new SolidColorBrush(Colors.Orange),
                StrokeThickness = 2,
                Fill = new SolidColorBrush(Colors.Orange),
                HorizontalAlignment = HorizontalAlignment.Stretch,
                VerticalAlignment = VerticalAlignment.Stretch,
                Width = 8,
                Height = 8
            };

            Rectangle rectangle = new Rectangle()
            {
                Fill = new SolidColorBrush(Colors.Orange),
                RadiusX = 2,
                RadiusY = 2,
                VerticalAlignment = VerticalAlignment.Stretch,
                HorizontalAlignment = HorizontalAlignment.Stretch,
                Height = 2
            };

            Grid.SetColumn(ellipse, 0);
            Grid.SetColumn(rectangle, 1);
            grid.Children.Add(ellipse);
            grid.Children.Add(rectangle);

            return grid;
        }

        //protected override void OnChanged()
        //{
        //    base.OnChanged();
        //}

        void AssociatedObject_RowLoaded(object sender, Telerik.Windows.Controls.GridView.RowLoadedEventArgs e)
        {
            if (e.Row is GridViewHeaderRow || e.Row is GridViewNewRow || e.Row is GridViewFooterRow)
                return;

            var row = e.Row as GridViewRow;
            this.InitializeRowDragAndDrop(row);
        }

        private void InitializeRowDragAndDrop(GridViewRow row)
        {
            if (row == null)
                return;

            row.SetValue(RadDragAndDropManager.AllowDragProperty, true);
            row.SetValue(RadDragAndDropManager.AllowDropProperty, true);

            RadDragAndDropManager.RemoveDropQueryHandler(row, OnGridViewRowDropQuery);
            RadDragAndDropManager.AddDropQueryHandler(row, OnGridViewRowDropQuery);
            RadDragAndDropManager.RemoveDropInfoHandler(row, OnGridViewRowDropInfo);
            RadDragAndDropManager.AddDropInfoHandler(row, OnGridViewRowDropInfo);
        }

        private void SubscribeToDragDropEvents()
        {
            RadDragAndDropManager.AddDropQueryHandler(this.AssociatedObject, OnDropQuery);
            RadDragAndDropManager.AddDropInfoHandler(this.AssociatedObject, OnDropInfo);
            RadDragAndDropManager.AddDragQueryHandler(this.AssociatedObject, OnDragQuery);
            RadDragAndDropManager.AddDragInfoHandler(this.AssociatedObject, OnDragInfo);
        }

        private void UnsubscribeFromDragDropEvents()
        {
            RadDragAndDropManager.RemoveDropQueryHandler(this.AssociatedObject, OnDropQuery);
            RadDragAndDropManager.RemoveDropInfoHandler(this.AssociatedObject, OnDropInfo);
            RadDragAndDropManager.RemoveDragQueryHandler(this.AssociatedObject, OnDragQuery);
            RadDragAndDropManager.RemoveDragInfoHandler(this.AssociatedObject, OnDragInfo);
        }

        public virtual DropPosition GetDropPositionFromPoint(Point absoluteMousePosition, GridViewRow row)
        {
            if (row != null)
            {
                var headerTopPoint = row.TransformToVisual(ApplicationHelper.RootVisual).Transform(new Point());
                double mouseTop = absoluteMousePosition.Y - headerTopPoint.Y;

                if (mouseTop <= (row.ActualHeight * 2 / 4))
                {
                    return DropPosition.Before;
                }
                else if (mouseTop > (row.ActualHeight * 2 / 4))
                {
                    return DropPosition.After;
                }
            }

            return DropPosition.Inside;
        }

        private void TryReorderRows(ReorderingEventArgs args)
        {
            IList source = args.SourceGrid.ItemsSource as IList;

            if (args.Handled || source == null)
                return;

            var newDropIndex = this.RemoveDraggedItems(source, args.DraggedItems, args.DropIndex);
            this.AddDraggedItems(source, args.DraggedItems, newDropIndex);
        }

        private int RemoveDraggedItems(IList sourceList, IEnumerable<object> draggedItems, int currentDropIndex)
        {
            var newDropIndex = UpdateDropIndex(draggedItems, sourceList, currentDropIndex);

            foreach (var itemToRemove in draggedItems)
                sourceList.Remove(itemToRemove);

            return newDropIndex;
        }

		private static int UpdateDropIndex(IEnumerable<object> draggedItems, IList sourceList, int currentDropIndex)
		{
            int newDropIndex = currentDropIndex;

			var itemsBeforeDropIndex =
			    from object item in draggedItems
			    where sourceList.IndexOf(item) < currentDropIndex
			    select item;

            foreach (var item in itemsBeforeDropIndex)
            {
                if (newDropIndex > 0) 
                    newDropIndex--;
            }

            return newDropIndex;
		}

        private void AddDraggedItems(IList sourceList, IEnumerable<object> draggedItems, int currentDropIndex)
        {
            for (int i = 0; i < draggedItems.Count(); i++)
            {
                sourceList.Insert(currentDropIndex + i, draggedItems.Skip(i).FirstOrDefault());
            }
        }

        protected void OnReordering(ReorderingEventArgs args)
        {
            if (this.Reordering != null)
                this.Reordering(this, args);
        }

        protected void OnReordered(ReorderedEventArgs args)
        {
            if (this.Reordered != null)
                this.Reordered(this, args);
        }

        private void OnDropQuery(object sender, DragDropQueryEventArgs e)
        {
            e.QueryResult = true;
            e.Handled = true;

            // CHECK IF THE SOURCE COLLECTION CAN BE MODIFIED
        }

        private void OnDropInfo(object sender, DragDropEventArgs e)
        {
            var gridView = e.Options.Destination as RadGridView;
            IEnumerable<object> draggedItems = e.Options.Payload as IEnumerable<object>;
            TreeViewDragCue cue = e.Options.DragCue as TreeViewDragCue;

            this.HideDropPositionFeedbackPresenter();

            if (e.Options.Status == DragStatus.DropPossible)
            {
                this.UpdateDragCueContent(cue, gridView.Items[gridView.Items.Count - 1], "Move after: ");
                cue.IsDropPossible = true;
            }
            else if (e.Options.Status == DragStatus.DropImpossible)
            {
                cue.DragActionContent = null;
                cue.IsDropPossible = false;
            }
            else if (e.Options.Status == DragStatus.DropComplete)
            {
                ReorderingEventArgs reorderingArgs = new ReorderingEventArgs(gridView, draggedItems, gridView.Items.Count);

                this.OnReordering(reorderingArgs);
                this.TryReorderRows(reorderingArgs);
                this.OnReordered(new ReorderedEventArgs(gridView, draggedItems));
            }

            e.Handled = true;
        }

        private void OnDragQuery(object sender, DragDropQueryEventArgs e)
        {
            var gridView = sender as RadGridView;

            if (gridView == null)
                return;

            e.QueryResult = this.InitiateDrag(gridView, e);
            e.Handled = true;
        }

        private void OnDragInfo(object sender, DragDropEventArgs e)
        {
            var gridView = sender as RadGridView;
            var draggedItems = e.Options.Payload as IEnumerable<object>;

            if (e.Options.Status == DragStatus.DragInProgress)
            {
                this.dragCue = this.CreateDragCue();
                this.dragCue.ItemsSource = draggedItems;

                e.Options.DragCue = this.dragCue;
            }
            else if (e.Options.Status == DragStatus.DragCancel)
            {
                //treeView.CancelDrag();
            }
            else if (e.Options.Status == DragStatus.DragComplete)
            {

            }
            else if (e.Options.Status == DragStatus.DropImpossible)
            {
                // HIDE DROP INDICATOR
            }

            e.Handled = true;
        }

        private void OnGridViewRowDropQuery(object sender, DragDropQueryEventArgs e)
        {
            var row = sender as GridViewRow;
            var draggedItems = e.Options.Payload as IEnumerable<object>;
            TreeViewDragCue cue = e.Options.DragCue as TreeViewDragCue;

            if (e.Options.Status == DragStatus.DropDestinationQuery)
            {
                this.currentDropItem = row.Item;
                this.currentDropPosition = this.GetDropPositionFromPoint(e.Options.CurrentDragPoint, row);

                e.QueryResult = !draggedItems.Contains(this.currentDropItem);

                if (!e.QueryResult.Value)
                {
                    this.ClearDragCueContent(cue);
                }
            }

            e.Handled = true;
        }

        private void OnGridViewRowDropInfo(object sender, DragDropEventArgs e)
        {
            var senderRow = (GridViewRow)sender;
            var gridView = senderRow.GridViewDataControl;
            IEnumerable<object> draggedItems = e.Options.Payload as IEnumerable<object>;

            TreeViewDragCue cue = e.Options.DragCue as TreeViewDragCue;

            if (e.Options.Status == DragStatus.DropPossible)
            {
                //var currentRow = this.AssociatedObject.ItemContainerGenerator.ContainerFromItem(this.currentDropItem) as GridViewRow;
                var currentRow = this.AssociatedObject.GetContainerFromDataItem((this.currentDropItem)) as GridViewRow;

                //if(currentRow == null)
                  // cancel drag

                this.UpdateDragCueContent(cue,
                    currentDropItem,
                    String.Format("Move {0}: ", Enum.GetName(typeof(DropPosition), this.currentDropPosition)).ToLower(CultureInfo.InvariantCulture));

                this.ShowDropPositionFeedbackPresenter(gridView, currentRow, this.currentDropPosition);
                cue.IsDropPossible = true;
            }
            else if (e.Options.Status == DragStatus.DropImpossible)
            {
                this.HideDropPositionFeedbackPresenter();

                this.ClearDragCueContent(cue);
                cue.IsDropPossible = false;

            }
            else if (e.Options.Status == DragStatus.DropComplete)
            {
                this.HideDropPositionFeedbackPresenter();

                var dropIndex = gridView.Items.IndexOf(this.currentDropItem);

                if (currentDropPosition == DropPosition.After)
                    dropIndex++;

                ReorderingEventArgs reorderingArgs = new ReorderingEventArgs(gridView, draggedItems, dropIndex);

                this.OnReordering(reorderingArgs);
                this.TryReorderRows(reorderingArgs);
                this.OnReordered(new ReorderedEventArgs(gridView, draggedItems));
            }

            e.Handled = true;
        }

        private bool IsDropPositionFeedbackAvailable()
        {
            return
                this.dropPositionFeedbackPresenterHost != null &&
                this.dropPositionFeedbackPresenter != null;
        }

        private void ShowDropPositionFeedbackPresenter(GridViewDataControl gridView, GridViewRow row, DropPosition lastRowDropPosition)
        {
            if (!this.IsDropPositionFeedbackAvailable())
                return;

            var yOffset = this.GetDropPositionFeedbackOffset(row, lastRowDropPosition);

            this.dropPositionFeedbackPresenter.Visibility = Visibility.Visible;
            this.dropPositionFeedbackPresenter.Width = row.ActualWidth;
            this.dropPositionFeedbackPresenter.RenderTransform = new TranslateTransform() { Y = yOffset };
        }

        private double GetDropPositionFeedbackOffset(GridViewRow row, DropPosition dropPosition)
        {
            var yOffset = row.TransformToVisual(this.dropPositionFeedbackPresenterHost).Transform(new Point(0, 0)).Y;

            if (dropPosition == DropPosition.After)
                yOffset += row.ActualHeight;

            yOffset -= (this.dropPositionFeedbackPresenter.ActualHeight / 2.0);

            return yOffset;
        }

        private void HideDropPositionFeedbackPresenter()
        {
            this.dropPositionFeedbackPresenter.RenderTransform = new TranslateTransform() { X = 0, Y = Double.NegativeInfinity };
        }

        private bool InitiateDrag(RadGridView gridView, DragDropQueryEventArgs dragDropArgs)
        {
            var itemsToDrag = this.GetItemsToReorder(gridView);
            var eventArgs = new BeginningDragEventArgs(gridView, itemsToDrag);

            this.OnBeginningDrag(eventArgs);

            if (!eventArgs.Cancel)
            {
                dragDropArgs.Options.Payload = itemsToDrag;
                this.OnDragStarted(new DragStartedEventArgs(gridView, itemsToDrag));
            }

            return !eventArgs.Cancel;
        }

        private IEnumerable<Object> GetItemsToReorder(RadGridView gridView)
        {
            if (this.PreserveSelectionOrderWhenReordering)
                return gridView.SelectedItems.ToList();
            else
            {
                List<object> itemsToReorder = new List<object>();

                for (int i = 0; i < gridView.Items.Count; i++)
                {
                    if (gridView.SelectedItems.Contains(gridView.Items[i]))
                        itemsToReorder.Add(gridView.Items[i]);
                }

                return itemsToReorder;
            }
        }

        protected virtual void OnBeginningDrag(BeginningDragEventArgs args)
        {
            if (this.BeginningDrag != null)
                this.BeginningDrag(this, args);
        }

        protected virtual void OnDragStarted(DragStartedEventArgs args)
        {
            if (this.DragStarted != null)
                this.DragStarted(this, args);
        }

        private TreeViewDragCue CreateDragCue()
        {
            var cue = new TreeViewDragCue();

            cue.Style = this.DragCueStyle;
            cue.DragTooltipContentTemplate = this.DragCueTooltipContentTemplate;
            cue.DragActionContentTemplate = this.DragCueActionContentTemplate;
            cue.ItemTemplate = this.DragCueItemTemplate;

            return cue;
        }

        private void UpdateDragCueContent(TreeViewDragCue cue, object tooltipContent, object actionContent)
        {
            cue.Style = this.DragCueStyle;
            cue.DragTooltipContentTemplate = this.DragCueTooltipContentTemplate;
            cue.DragActionContentTemplate = this.DragCueActionContentTemplate;
            cue.ItemTemplate = this.DragCueItemTemplate;

            cue.DragTooltipContent = tooltipContent;
            cue.DragActionContent = actionContent;
        }

        private void ClearDragCueContent(TreeViewDragCue cue)
        {
            cue.DragTooltipContentTemplate = null;
            cue.DragActionContentTemplate = null;
            cue.DragTooltipContent = null;
            cue.DragActionContent = null;
        }
    }
}
